/******************************************************************************
 *  Author:		Joey Richey (CM 1698)                                         *
 *  Project:	Maze Generator                                                *
 *	Filename:	maze.cpp (main source file. does all the windowing,           *
 *				  displaying, and calculates the walls					      *
 *                                                                            *
 *  This program generates a random maze. Every square is reachable from      *
 *  every other square. It can be any size (square or rectangle). It is       *
 *  only limited by screen resolution and  memory). After the maze is         *
 *	generated, it is also solved. after it is solved the user may press       *
 *	spacebar to see the solution to the maze. For best results, width and     *
 *  height should be relatively close (not more than a factor of 4 difference)*
 *                                                                            *
 *  Changelog                                                                 *
 *		9-12-02	created, allows random mazes of various sizes to be created   *
 *		9-14-02	added thread support (through zthreads or pthreads for the    *
 *				solver script (not yet made), also have nonthread version     *
 *		9-15-02	made the solver "function" and had it display solution		  *
 *				                                                              *
 *                                                                            *
 ******************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <time.h>		/*	random numbers				*/
#include <GL/glut.h>	/*	needed for graphics			*/
#include "defines.h"	/*	#define we will need		*/
#include "union.h"		/*	my ghetto Union structure	*/
#include "solveMaze.h"	/*	solves the maze for us		*/

/*=============================
	function declarations	
===============================*/

/*	Gets input from the user	*/
void getInput(void);
/*	inits data to have all the walls between the cells setup correctly	*/
void generateMazeWalls(void);
/*	swaps 2 integers	*/
void swap(int *, int *);
/*	puts numbers between 0..max-1 into retArray in a random order, retArray 
	must be initlized to max size	*/
int randList(int retArray[], int max);
/*	removes walls so the maze is acutally playable	*/
void removeWalls(void);
/*	inits the graphics data for the window	*/
void gfxinit (void);
/*	loopback function for displaying in glut	*/
void display (void);
/*	loopback function for handling input	*/
void handleInput (unsigned char, int, int);


/*==============================
	global variable
==============================*/

/*	as much as i tried, i am stuck using a couple of global variables		*/
bool mazeSolved = false;		/*	whether we have solved the maze yet		*/
MUTEX_CREATE(mutex_mazeSolved);	/*	must protect shared memory				*/
boardData board;				/*	info about the maze						*/
wallData *walls;				/*	stores info about the walls				*/
int *solvedPath;				/*	an array that will hold the fastest 
									solved path that we can then draw with	*/
int fastestPathLength;			/*	least number of moves to get to the end	*/
bool displayMazeSolution;		/*	whether we should draw the soltion to 
									the maze							 	*/


/*======================================
	main actually begins here
======================================*/

int main(int argc, char** argv)
{
	
	int i;							/*	loopindex	*/
	int numWalls;					/*	how many interior wall spaces there 
										are	*/
	int numCells;					/*	number of cells in the maze	*/
	int screenWidth, screenHeight;	/*	stores the width/height of screen	*/
	int windowWidth, windowHeight;	/*	stores the width/height of the window*/
	float widthScale, heightScale;	/*	how much to scale the window if
										rectangular board	*/
	extern boardData board;
	extern wallData *walls;

	/*	we don't want to display a solution before we have one	*/
	displayMazeSolution = false;

	/*	cross platform	init the random number generator	*/
	SEED ((int) time(NULL));
	
	/*	get input	*/
	getInput();
	printf("Generating maze...\n");
	
	/*	now we will have to geneate a list for all the "walls" to be in
		it will contain (height*(width-1)) + (height-1)*width items */
	
	numWalls = (board.height*(board.width-1)) + (board.height-1)*board.width;
	numCells = board.height*board.width;

	walls = (wallData *) calloc(numWalls, sizeof(wallData));
	if (walls == NULL)
		fatalError(CALLOC_MEM);

	/*	initlize the contents of walls	*/
	for (i=0;i<numWalls;i++)
	{
		walls[i].first=-1;
		walls[i].last=-1;
		walls[i].hasWall=true;
	}

#ifdef DEBUG
	printf("Done generating (%d) walls\n",numWalls);
#endif
	
	/*	now that we have memory for the walls, lets make them do something 
		useful	*/
	generateMazeWalls();

	/*	now so the maze is actually worthwhile, let's remove some walls and 
		have every square connected to every other square	*/

	removeWalls();

	printf("Done generating maze...\n");
#ifdef DONT_USE_THREADS
	/*	if it's windows and we don't have zthreads we have to solve all the
		maze before hand	*/
	solveMaze(NULL);
#endif

	/*	now that we have the data for the maze, let's setup some glut stuff
		so we can draw it	*/

	/*	lets get some values to make the window look pretty	*/
	
	/*	the SGIs segfault if you try to get the screen width	*/
#ifdef _WIN32
	screenWidth = glutGet(GLUT_SCREEN_WIDTH);
	screenHeight = glutGet(GLUT_SCREEN_HEIGHT);
#else
	/*	trial and error to get screen size for SGIs	*/
	screenWidth = SCREEN_WIDTH;
	screenHeight = SCREEN_HEIGHT;
#endif

	if ((screenWidth == 0) || (screenHeight == 0))
		fatalError(RETRIEVE_SCREEN);

	/*	the size of the window, if rectangle, scale it so the squares do not
		get all distorted	*/
	if(board.width > board.height)
	{
		widthScale = 1.0;
		heightScale = (float)board.height/(float)board.width;
	}
	else if(board.width < board.height)
	{
		heightScale = 1.0;
		widthScale = (float)board.width/(float)board.height;
	}
	else	/*	==	*/
	{
		widthScale = 1.0;
		heightScale = 1.0;
	}

	windowWidth = PERCENT_FILL_SCREEN * screenWidth * widthScale;
	windowHeight = PERCENT_FILL_SCREEN * screenHeight * heightScale;


	glutInit (&argc, argv);
	glutInitWindowSize (windowWidth, windowHeight);
	/*	center the window	*/
	glutInitWindowPosition ((screenWidth - windowWidth) * 0.5, 
							(screenHeight-windowHeight) * 0.5);
	glutInitDisplayMode (GLUT_SINGLE | GLUT_RGB);
	glutCreateWindow ("Maze");
	glutDisplayFunc (display);
	glutKeyboardFunc(handleInput);
	gfxinit ();


#ifdef DONT_USE_THREADS
	/*	if we aren't using threads, we no longer need this info, so let's
		free it so we aren't taking up that memory during execution	*/
	free(walls);
#else
	/*	start the thread that will solve the maze	*/
	THREAD_CREATE(thread, attr);
	THREAD_INIT(attr);
	THREAD_RUN(thread,attr,solveMaze);
#endif

	glutMainLoop ();
	return 0;
}


void removeWalls(void)
{
	/*	
		this procedure will remove walls from the interior of the maze so it is 
		actually playable every square will be able to be accessed from any 
		other square
		The algorithim for doing this is taken from pages 803-806 in "Data 
		Structres & Problem Solving Using Java" by Mark Weiss
	*/
	/*	
		we need to select adjacent cells at random. To accomplish this, i will
		get an array that contains numbers in a random order. Then i will 
		access each element of that array sequentially, taking the random value
		it holds to access values in wall in a random order
	*/

	int *wallOrder;		/*	the order which we will go thorugh walls	*/
	int first,last;		/*	vars to speed things up	*/
	int i;				/*	loopindex	*/
	extern boardData board;
	extern wallData *walls;
	int numWalls = (board.height*(board.width-1)) + (board.height-1)*board.width;
	int numCells = board.height*board.width;


	wallOrder = (int *) calloc(numWalls,sizeof(int));
	if(wallOrder == NULL)
		fatalError(CALLOC_MEM);

	if(randList(wallOrder,numWalls) < 0)
		fatalError(RAND_LIST);

	/*	now that we get there, we have to setup some kind of union structure	*/
	Union DestoryWall(numCells);
	for (i=0;i<numCells;i++)
		if (DestoryWall.Create(i) < 0)
			fatalError(UNION_CREATE);


#ifdef DEBUG
	printf("Unions succesffuly created (%d) of them\n", numCells);
#endif

	for(i=0;i<numWalls;i++)
	{
		first = walls[wallOrder[i]].first;
		last = walls[wallOrder[i]].last;

		if(!DestoryWall.Matches(first,last))
		{
			/*	these two points do not connect, destory the wall between them 
				and put them in the same union	*/
			walls[wallOrder[i]].hasWall = false;
			DestoryWall.Merge(last, first);
#ifdef DEBUG
			printf("removed wall between squares %d:(%d, %d)\n",wallOrder[i],first,last);
#endif
		}
	}

	/*	free some memory before we return	*/
	free(wallOrder);

}

void gfxinit (void)
/* This is the routine that generates the image to be displayed. */
{
	int i;								/*	loop counter	*/
	int numWalls; 
 	float x1,y1,x2,y2;					/*	speed up calculations	*/
	extern boardData board;
	extern wallData *walls;

	/*	calculate some variables we will use later	*/
	numWalls = (board.height*(board.width-1))+(board.height-1)*board.width;


	glClearColor (1.0, 1.0, 1.0, 0.0); /* Make the background white.         */
	glColor3f (0.0, 0.0, 0.0);         /* Draw in black.                     */
	glMatrixMode (GL_PROJECTION);
	glLoadIdentity ();
	gluOrtho2D (-1, board.width + 1 , board.height + 1, -1);

	glNewList (MAZE, GL_COMPILE);
		/*	draw the square around the board	*/
		glBegin(GL_LINE_STRIP);
			glVertex2f(1.0,0.0);
			glVertex2f(board.width, 0.0);
			glVertex2f(board.width,board.height-1);
		glEnd();
		glBegin(GL_LINE_STRIP);
			glVertex2f(0.0,0.0);
			glVertex2f(0.0, board.height);
			glVertex2f(board.width,board.height);
		glEnd();

		/*	now let's actually draw the board data the we calculated before	*/
#ifndef HAVE_GHETTO_VIDEO_CARD
		glBegin(GL_LINES);
#endif
			/*	this will put a lot of verticies through twice, but I don't 
				really think there is a better way to do this since I don't 
				know which ones	will be duplicated	*/
			for(i=0;i<numWalls;i++)
			{
				if(walls[i].hasWall == true)
				{
					/*	must determine if wall is verticle or horizontle	*/
					if (walls[i].first + 1 == walls[i].last)
					{
						/*	Verticle	*/
						x1 = x2 = (walls[i].first % board.width + 1);
						y1 = (walls[i].first / board.width);
						y2 = y1 + 1;
#ifdef DEBUG
						printf("Vertical:");
#endif

					}
					else
					{
						/*	horizontel	*/
						y1 = y2 = (walls[i].first / board.width + 1);
						x1 = (walls[i].first % board.width);
						x2 = x1 + 1;
#ifdef DEBUG
						printf("Horizontal:");
#endif
					}
#ifdef HAVE_GHETTO_VIDEO_CARD
					glBegin(GL_LINES);
#endif
					glVertex2f(x1,y1);
					glVertex2f(x2,y2);
#ifdef HAVE_GHETTO_VIDEO_CARD
					glEnd();
#endif
#ifdef DEBUG
					printf("drew a maze line %d: (%f,  %f, %f, %f)(%d,%d)\n",
							i, x1,y1,x2,y2,walls[i].first,walls[i].last);
#endif
				}
			}
#ifndef	HAVE_GHETTO_VIDEO_CARD
		glEnd();
#endif
	glEndList();
}


void display (void)
/* This is the callback function that gets executed every time the display
   needs to be updated.
*/
{
	extern bool mazeSolved;
	extern int fastedPathLength;
	extern int* soldData;
	extern bool displayMazeSolution;
	static bool mazeDrawn = false;
	int i;
	float x2Use, y2Use;

#ifndef DONT_USE_THREADS
	/*	todo:high:this may messup under pthreads	*/
	extern MUTEX_CREATE(mutex_mazeSolved);
#endif

	glClear (GL_COLOR_BUFFER_BIT);
	glColor3f (0.0, 0.0, 0.0);         /*	Draw in red	*/
	glCallList (MAZE);
	
	/*	inorder to get drawing right, I have to update the info for the solved
		maze in display and not in the thread. If the maze is solved and not
		yet drawn, draw it
	*/

	MUTEX_LOCK(mutex_mazeSolved);
	if (mazeSolved)
	{
		MUTEX_UNLOCK(mutex_mazeSolved);	
		if (!mazeDrawn)
		{
			/*	now that we have a path to solve the maze, let's do the gl 
				stuff so that we can have it drawn	*/
			glNewList (MAZE_SOLUTION, GL_COMPILE);
				glBegin(GL_LINE_STRIP);
					/*	the solution is in reverse order, so start from lower
						right	*/
					glVertex2f((float)board.width-0.5, (float)board.height-0.5);
					for(i=0;i<fastestPathLength;i++)
					{
						y2Use = (float)(solvedPath[i] / (board.width)) + 0.5;
						x2Use = (float)(solvedPath[i] % (board.width)) + 0.5;
						glVertex2f(x2Use, y2Use);
					}
				glEnd();

			glEndList();
			free(solvedPath);
			mazeDrawn = true;
		}
		/*	only show the maze if the user wants to see it	*/
		if (displayMazeSolution)
		{
			glColor3f (1.0, 0.0, 0.0);         /*	Draw in red	*/
			glCallList(MAZE_SOLUTION);
		}
	}
	else
		MUTEX_UNLOCK(mutex_mazeSolved);
	
	
	glFlush ();
}




/*	Gets the input needed from the user	*/
void getInput()
{
	extern boardData board;

	printf("< Maze Generator >\n\n");
	printf("Reasonalbe values for height and width are < 150, but feel free to go higher\n\n");
	printf("Please enter the width of the maze: ");
	fscanf(stdin,"%d", &board.width);
	printf("Please enter the height of the maze: ");
	fscanf(stdin,"%d", &board.height);
#ifdef DEBUG
	printf("Board width:\t%d\nBoard height:\t%d\n",board.width, board.height);
#endif
}

void fatalError(int err)
{
	fprintf(stderr,"A fatal error has occured in the program and it is \
					exiting with level %d\n",err);
	exit(err);

}


void generateMazeWalls(void)
{
	int i,j;		/*	loopindex	*/
	int tmp;		/*	temp variables to speed up repeated calculations.	*/
	int index=-1;	/*	keep track of where in the array we are	*/
	extern boardData board;
	extern wallData *walls;

	for(i=0;i<board.height;i++)
	{
		for(j=0;j<board.width;j++)
		{
			tmp = i*board.width+j;
			/*	the wall the goes to the right and the last column doesn't 
				have one	*/
			if(j != (board.width -1))
			{
				walls[++index].first=tmp;
				walls[index].last=tmp+1;
#ifdef DEBUG
				printf("Added wall(1) at index %d with walls(%d  ,%d)\n",
						index,walls[index].first,walls[index].last);
#endif
			}

			/*	the wall that does down (but not the last row)	*/
			if( i != (board.height-1))
			{
				walls[++index].first=tmp;
				walls[index].last=(i+1)*board.width+j;
#ifdef DEBUG
				printf("Added wall(2) at index %d with walls(%d  ,%d)\n",
						index,walls[index].first,walls[index].last);
#endif
			}
		}
	}

}

void handleInput (unsigned char key, int x, int y)
{
	extern bool displayMazeSolution;
	extern bool mazeSolved;
	if ((key == 32) && (mazeSolved))
	{
		displayMazeSolution = !displayMazeSolution;
		/*	forece refresh	*/
		glutPostRedisplay();
	}

}
/* begin Maintenance functions	*/
void swap(int *a, int *b)
{
	int tmp=*a;
	*a=*b;
	*b=tmp;
}


int randList(int retArray[], int max)
{
/*	I'm not a math major, but this seems to generate 
	a pseudo random list of numbers	*/

	int i, temp;
	int *tmpArray;
	int indexToUse=max;

	if (max==0)
	{
		retArray[0] = 0;
		return 0;
	}
	tmpArray = (int *) calloc(max, sizeof(int));
	if (tmpArray == NULL)
	{
		/*	error allocating memory	*/
		return -1;
	}
	for (i=0;i<max;i++)
		tmpArray[i]=i;
	for (i=0;i<max;i++)
	{
		temp = (RANDNUM() % (indexToUse))+1;			
		retArray[i] = tmpArray[temp-1];
		swap(&tmpArray[--temp],&tmpArray[--indexToUse]);
	}
	free(tmpArray);
	return 0;
}
